from graph_retriever.testing.embeddings import AnimalEmbeddings
from langchain_core.vectorstores.in_memory import InMemoryVectorStore
from langchain_graph_retriever import GraphRetriever
from langchain_tests.integration_tests import RetrieversIntegrationTests

from tests.conftest import load_animal_docs
from tests.embeddings import BaseEmbeddings


class TestGraphTraversalRetriever(RetrieversIntegrationTests):
    @property
    def retriever_constructor(self) -> type[GraphRetriever]:
        return GraphRetriever

    @property
    def retriever_constructor_params(self) -> dict:
        embedding = BaseEmbeddings(AnimalEmbeddings())
        store = InMemoryVectorStore(embedding=embedding)
        store.add_documents(load_animal_docs())
        return {
            "store": store,
            "edges": [("habitat", "habitat")],
        }

    @property
    def retriever_query_example(self) -> str:
        """
        Returns a str representing the "query" of an example retriever call.
        """
        return "what are some small agile mammals?"



"""Infers the appropriate adapter for a given vector store."""

import importlib

from graph_retriever.adapters import Adapter
from langchain_core.vectorstores.base import VectorStore

ADAPTERS_PKG = "langchain_graph_retriever.adapters"
_KNOWN_STORES = {
    "langchain_astradb.vectorstores.AstraDBVectorStore": (
        f"{ADAPTERS_PKG}.astra",
        "AstraAdapter",
    ),
    "langchain_community.vectorstores.cassandra.Cassandra": (
        f"{ADAPTERS_PKG}.cassandra",
        "CassandraAdapter",
    ),
    "langchain_chroma.vectorstores.Chroma": (f"{ADAPTERS_PKG}.chroma", "ChromaAdapter"),
    "langchain_core.vectorstores.in_memory.InMemoryVectorStore": (
        f"{ADAPTERS_PKG}.in_memory",
        "InMemoryAdapter",
    ),
    (
        "langchain_community.vectorstores."
        "opensearch_vector_search.OpenSearchVectorSearch"
    ): (
        f"{ADAPTERS_PKG}.open_search",
        "OpenSearchAdapter",
    ),
}

# Class names that indicate we don't need to keep traversing.
STOP_NAMES = {
    "abc.ABC",
    "builtins.object",
    "langchain_core.vectorstores.base.VectorStore",
}


def _full_class_name(cls: type) -> str:
    return f"{cls.__module__}.{cls.__name__}"


def _infer_adapter_name(cls: type) -> tuple[str, str]:
    """Return the module and class of the adapter or raise."""
    store_classes = [cls]
    while store_classes:
        store_class = store_classes.pop()

        store_class_name = _full_class_name(store_class)
        if store_class_name in STOP_NAMES:
            continue

        adapter = _KNOWN_STORES.get(store_class_name, None)
        if adapter is not None:
            return adapter

        # If we didn't find it yet, and the naem wasn't a stopping point,
        # we queue up the base classes for consideration. This allows
        # matching subclasses of supported vector stores.
        store_classes.extend(store_class.__bases__)

    store_class_name = _full_class_name(cls)
    raise ValueError(
        f"Expected adapter or supported vector store, but got {store_class_name}"
    )


def infer_adapter(store: Adapter | VectorStore) -> Adapter:
    """
    Dynamically infer the adapter for a given vector store.

    This function identifies the correct adapter based on the vector store type
    and instantiates it with the provided arguments.

    Parameters
    ----------
    store :
        The vector store instance.

    Returns
    -------
    :
        The initialized adapter for the given vector store.
    """
    if isinstance(store, Adapter):
        return store

    module_name, class_name = _infer_adapter_name(store.__class__)
    adapter_module = importlib.import_module(module_name)
    adapter_class = getattr(adapter_module, class_name)
    return adapter_class(store)



"""Provides an adapter for OpenSearch vector store integration."""

from collections.abc import Sequence
from typing import Any

from typing_extensions import override

try:
    from langchain_community.vectorstores.opensearch_vector_search import (
        OpenSearchVectorSearch,
    )
except (ImportError, ModuleNotFoundError):
    raise ImportError("please `pip install langchain-community opensearch-py`")

from langchain_core.documents import Document

from langchain_graph_retriever._conversion import METADATA_EMBEDDING_KEY
from langchain_graph_retriever.adapters.langchain import LangchainAdapter


class OpenSearchAdapter(LangchainAdapter[OpenSearchVectorSearch]):
    """
    Adapter to traverse OpenSearch vector stores.

    This adapter enables similarity search and document retrieval using an
    OpenSearch vector store.

    Parameters
    ----------
    vector_store  :
        The OpenSearch vector store instance.

    Notes
    -----
    Graph Traversal is only supported when using either the `"lucene"` or
    `"faiss"` engine.

    For more info, see the [OpenSearch Documentation](https://opensearch.org/docs/latest/search-plugins/knn/knn-index#method-definitions)
    """

    def __init__(self, vector_store: OpenSearchVectorSearch):
        if vector_store.engine not in ["lucene", "faiss"]:
            msg = (
                f"Invalid engine for Traversal: '{self.vector_store.engine}'"
                " please instantiate the Open Search Vector Store with"
                " either the 'lucene' or 'faiss' engine"
            )
            raise ValueError(msg)
        super().__init__(vector_store)

        if vector_store.is_aoss:
            self._id_filed = "id"
        else:
            self._id_field = "_id"

    def _build_filter(
        self, filter: dict[str, str] | None = None
    ) -> list[dict[str, Any]] | None:
        """
        Build a filter query for OpenSearch based on metadata.

        Parameters
        ----------
        filter :
            Metadata filter to apply.

        Returns
        -------
        :
            Filter query for OpenSearch.
        """
        if filter is None:
            return None
        return [
            {
                "terms" if isinstance(value, list) else "term": {
                    f"metadata.{key}.keyword": value
                }
            }
            for key, value in filter.items()
        ]

    @override
    def _search(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        if filter is not None:
            # use an efficient_filter to collect results that
            # are near the embedding vector until up to 'k'
            # documents that match the filter are found.
            kwargs["efficient_filter"] = {
                "bool": {"must": self._build_filter(filter=filter)}
            }

        if k == 0:
            return []

        docs = self.vector_store.similarity_search_by_vector(
            embedding=embedding,
            k=k,
            metadata_field="*",
            **kwargs,
        )

        # if metadata=="*" on the search, then the document
        # embedding vector and text are included in the
        # document metadata in the returned document.
        #
        # The actual document metadata is moved down into a
        # sub "metadata" key.
        return [
            Document(
                id=doc.id,
                page_content=doc.page_content,
                metadata={
                    METADATA_EMBEDDING_KEY: doc.metadata["vector_field"],
                    **doc.metadata["metadata"],
                },
            )
            for doc in docs
        ]

    @override
    def _get(
        self, ids: Sequence[str], filter: dict[str, Any] | None = None, **kwargs: Any
    ) -> list[Document]:
        query: dict[str, Any] = {"ids": {"values": ids}}

        if filter:
            query = {
                "bool": {"must": [query, *(self._build_filter(filter=filter) or [])]}
            }

        response = self.vector_store.client.search(
            body={
                "query": query,
            },
            index=self.vector_store.index_name,
            _source_includes=["text", "metadata", "vector_field"],
            **kwargs,
        )

        return [
            Document(
                page_content=hit["_source"]["text"],
                metadata={
                    METADATA_EMBEDDING_KEY: hit["_source"]["vector_field"],
                    **hit["_source"]["metadata"],
                },
                id=hit["_id"],
            )
            for hit in response["hits"]["hits"]
        ]



"""Provides an adapter for AstraDB vector store integration."""

from __future__ import annotations

import asyncio
import warnings
from collections.abc import Iterable, Sequence
from typing import Any, cast

import backoff
from graph_retriever import Content
from graph_retriever.edges import Edge, IdEdge, MetadataEdge
from graph_retriever.utils.batched import batched
from graph_retriever.utils.top_k import top_k
from typing_extensions import override

try:
    from langchain_astradb import AstraDBVectorStore
    from langchain_astradb.utils.vector_store_codecs import (
        _AstraDBVectorStoreDocumentCodec,
    )
except (ImportError, ModuleNotFoundError):
    raise ImportError("please `pip install langchain-astradb`")

try:
    import astrapy
except (ImportError, ModuleNotFoundError):
    raise ImportError("please `pip install astrapy")
import httpx
from graph_retriever.adapters import Adapter
from langchain_core.documents import Document

_EXCEPTIONS_TO_RETRY = (
    httpx.TransportError,
    astrapy.exceptions.DataAPIException,
)
_MAX_RETRIES = 3


class _QueryHelper:
    def __init__(
        self,
        codec: _AstraDBVectorStoreDocumentCodec,
        user_filter: dict[str, Any] | None = {},
    ) -> None:
        self.codec = codec
        self.encoded_user_filter = (
            codec.encode_filter(user_filter) if user_filter else {}
        )

    def encode_filter(self, filter: dict[str, Any]) -> dict[str, Any]:
        encoded_filter = self.codec.encode_filter(filter)
        return self._with_user_filter(encoded_filter)

    def _with_user_filter(self, encoded_filter: dict[str, Any]) -> dict[str, Any]:
        if self.encoded_user_filter:
            return {"$and": [encoded_filter, self.encoded_user_filter]}
        else:
            return encoded_filter

    def create_ids_query(self, ids: list[str]) -> dict[str, Any]:
        if len(ids) == 0:
            raise ValueError("IDs should not be empty")
        elif len(ids) == 1:
            # We need to avoid re-encoding in this case, since the codecs
            # don't recognize `_id` and get confused (creating `metadata._id`).
            return self._with_user_filter({"_id": ids[0]})
        elif len(ids) < 100:
            return self._with_user_filter({"_id": {"$in": ids}})
        else:
            raise ValueError("IDs should be less than 100, was {len(ids)}")

    def create_metadata_query(
        self, metadata: dict[str, Iterable[Any]]
    ) -> dict[str, Any] | None:
        metadata = {k: v for k, v in metadata.items() if v}
        if not metadata:
            return None

        parts = []
        for k, v in metadata.items():
            # If there are more than 100 values, we can't create a single `$in` query.
            # But, we can do it for each batch of 100.
            for v_batch in batched(v, 100):
                batch = list(v_batch)
                if len(batch) == 1:
                    parts.append({k: batch[0]})
                else:
                    parts.append({k: {"$in": batch}})

        if len(parts) == 1:
            return self.encode_filter(parts[0])
        else:
            return self.encode_filter({"$or": parts})


class AstraAdapter(Adapter):
    """
    Adapter for the [AstraDB](https://www.datastax.com/products/datastax-astra) vector store.

    This class integrates the LangChain AstraDB vector store with the graph
    retriever system, providing functionality for similarity search and document
    retrieval.

    Parameters
    ----------
    vector_store :
        The AstraDB vector store instance.
    """  # noqa: E501

    def __init__(self, vector_store: AstraDBVectorStore) -> None:
        self.vector_store = vector_store.copy(
            component_name="langchain_graph_retriever"
        )

    def _build_contents(
        self, docs_with_embeddings: list[tuple[Document, list[float]]]
    ) -> list[Content]:
        contents = []
        for doc, embedding in docs_with_embeddings:
            assert doc.id is not None
            contents.append(
                Content(
                    id=doc.id,
                    content=doc.page_content,
                    metadata=doc.metadata,
                    embedding=embedding,
                )
            )
        return contents

    @override
    @backoff.on_exception(backoff.expo, _EXCEPTIONS_TO_RETRY, max_tries=_MAX_RETRIES)
    def search_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> tuple[list[float], list[Content]]:
        # Work around the fact that `k == 0` is rejected by Astra.
        # AstraDBVectorStore has a similar work around for non-vectorize path, but
        # we want it to apply in both cases.
        query_k = k
        if query_k == 0:
            query_k = 1
        query_embedding, docs_with_embeddings = (
            self.vector_store.similarity_search_with_embedding(
                query=query,
                k=query_k,
                filter=filter,
                **kwargs,
            )
        )
        if k == 0:
            return query_embedding, []

        return query_embedding, self._build_contents(docs_with_embeddings)

    @override
    @backoff.on_exception(backoff.expo, _EXCEPTIONS_TO_RETRY, max_tries=_MAX_RETRIES)
    async def asearch_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> tuple[list[float], list[Content]]:
        # Work around the fact that `k == 0` is rejected by Astra.
        # AstraDBVectorStore has a similar work around for non-vectorize path, but
        # we want it to apply in both cases.
        query_k = k
        if query_k == 0:
            query_k = 1
        (
            query_embedding,
            docs_with_embeddings,
        ) = await self.vector_store.asimilarity_search_with_embedding(
            query=query,
            k=query_k,
            filter=filter,
            **kwargs,
        )
        if k == 0:
            return query_embedding, []
        return query_embedding, self._build_contents(docs_with_embeddings)

    @override
    @backoff.on_exception(backoff.expo, _EXCEPTIONS_TO_RETRY, max_tries=_MAX_RETRIES)
    def search(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        if k == 0:
            return []

        docs_with_embeddings = (
            self.vector_store.similarity_search_with_embedding_by_vector(
                embedding=embedding,
                k=k,
                filter=filter,
                **kwargs,
            )
        )
        return self._build_contents(docs_with_embeddings)

    @override
    @backoff.on_exception(backoff.expo, _EXCEPTIONS_TO_RETRY, max_tries=_MAX_RETRIES)
    async def asearch(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        if k == 0:
            return []

        docs_with_embeddings = (
            await self.vector_store.asimilarity_search_with_embedding_by_vector(
                embedding=embedding,
                k=k,
                filter=filter,
                **kwargs,
            )
        )
        return self._build_contents(docs_with_embeddings)

    @override
    def get(
        self, ids: Sequence[str], filter: dict[str, Any] | None = None, **kwargs: Any
    ) -> list[Content]:
        helper = _QueryHelper(self.vector_store.document_codec, filter)

        results: dict[str, Content] = {}
        for batch in batched(set(ids), 100):
            query = helper.create_ids_query(list(batch))

            # TODO: Consider deduplicating before decoding?
            for content in self._execute_query(query=query):
                results.setdefault(content.id, content)
        return list(results.values())

    @override
    async def aget(
        self, ids: Sequence[str], filter: dict[str, Any] | None = None, **kwargs: Any
    ) -> list[Content]:
        helper = _QueryHelper(self.vector_store.document_codec, filter)

        tasks = set()
        for batch in batched(set(ids), 100):
            query = helper.create_ids_query(list(batch))
            tasks.add(asyncio.create_task(self._aexecute_query(query=query)))

        results: dict[str, Content] = {}
        while tasks:
            done, pending = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED
            )
            tasks = pending

            # TODO: Consider deduplicating before decoding?
            for contents in done:
                for content in await contents:
                    results.setdefault(content.id, content)

        return list(results.values())

    def _prepare_edge_query_parts(
        self, edges: set[Edge]
    ) -> tuple[dict[str, Iterable[Any]], set[str]]:
        """
        Return metadata and ID query parts for edges.

        Parameters
        ----------
        edges :
            The edges to prepare.

        Returns
        -------
        metadata_in :
            Dictionary of metadata constraints indicating the field to constrain
            and the set of values.
        ids :
            Set of IDs to query for.

        Raises
        ------
        ValueError
            If any edges are invalid.
        """
        metadata_in: dict[str, set[Any]] = {}
        ids = set()

        for edge in edges:
            if isinstance(edge, MetadataEdge):
                metadata_in.setdefault(edge.incoming_field, set()).add(edge.value)
            elif isinstance(edge, IdEdge):
                ids.add(edge.id)
            else:
                raise ValueError(f"Unsupported edge {edge}")

        return (cast(dict[str, Iterable[Any]], metadata_in), ids)

    @override
    def adjacent(
        self,
        edges: set[Edge],
        query_embedding: list[float],
        k: int,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ) -> Iterable[Content]:
        query_helper = _QueryHelper(self.vector_store.document_codec, filter)
        metadata, ids = self._prepare_edge_query_parts(edges)

        sort = self.vector_store.document_codec.encode_vector_sort(query_embedding)

        results = []
        if (metadata_query := query_helper.create_metadata_query(metadata)) is not None:
            batch = self._execute_query(
                limit=k,
                sort=sort,
                query=metadata_query,
            )
            results.extend(batch)
            ids.difference_update({c.id for c in batch})

        for id_batch in batched(ids, 100):
            results.extend(
                self._execute_query(
                    limit=k,
                    sort=sort,
                    query=query_helper.create_ids_query(list(id_batch)),
                )
            )

        return top_k(results, embedding=query_embedding, k=k)

    @override
    async def aadjacent(
        self,
        edges: set[Edge],
        query_embedding: list[float],
        k: int,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ) -> Iterable[Content]:
        query_helper = _QueryHelper(self.vector_store.document_codec, filter)
        metadata, ids = self._prepare_edge_query_parts(edges)

        sort = self.vector_store.document_codec.encode_vector_sort(query_embedding)

        results = []
        if (metadata_query := query_helper.create_metadata_query(metadata)) is not None:
            batch = await self._aexecute_query(
                limit=k,
                sort=sort,
                query=metadata_query,
            )
            results.extend(batch)
            ids.difference_update({c.id for c in batch})

        for id_batch in batched(ids, 100):
            # I have tested (in the lazy graph rag example) doing these
            # two queries concurrently, but it hurt perfomance. Likely this
            # would depend on how many IDs the previous query eliminates from
            # this query.
            result_batch = await self._aexecute_query(
                limit=k,
                sort=sort,
                query=query_helper.create_ids_query(list(id_batch)),
            )
            results.extend(result_batch)

        return top_k(results, embedding=query_embedding, k=k)

    def _execute_query(
        self,
        query: dict[str, Any] | None = None,
        limit: int | None = None,
        sort: dict[str, Any] | None = None,
    ) -> list[Content]:
        astra_env = self.vector_store.astra_env
        astra_env.ensure_db_setup()

        # Similarity can only be included if we are sorting by vector.
        # Ideally, we could request the `$similarity` projection even
        # without vector sort. And it can't be `False`. It needs to be
        # `None` or it will cause an assertion error.
        include_similarity = None
        if not (sort or {}).keys().isdisjoint({"$vector", "$vectorize"}):
            include_similarity = True
        hits = astra_env.collection.find(
            filter=query,
            projection=self.vector_store.document_codec.full_projection,
            limit=limit,
            include_sort_vector=True,
            include_similarity=include_similarity,
            sort=sort,
        )

        return [content for hit in hits if (content := self._decode_hit(hit))]

    async def _aexecute_query(
        self,
        query: dict[str, Any] | None = None,
        limit: int | None = None,
        sort: dict[str, Any] | None = None,
    ) -> list[Content]:
        astra_env = self.vector_store.astra_env
        await astra_env.aensure_db_setup()

        # Similarity can only be included if we are sorting by vector.
        # Ideally, we could request the `$similarity` projection even
        # without vector sort. And it can't be `False`. It needs to be
        # `None` or it will cause an assertion error.
        include_similarity = None
        if not (sort or {}).keys().isdisjoint({"$vector", "$vectorize"}):
            include_similarity = True
        hits = astra_env.async_collection.find(
            filter=query,
            projection=self.vector_store.document_codec.full_projection,
            limit=limit,
            include_sort_vector=True,
            include_similarity=include_similarity,
            sort=sort,
        )

        return [content async for hit in hits if (content := self._decode_hit(hit))]

    def _decode_hit(self, hit: dict[str, Any]) -> Content | None:
        codec = self.vector_store.document_codec
        if "metadata" not in hit or codec.content_field not in hit:
            id = hit.get("_id", "(no _id)")
            warnings.warn(
                f"Ignoring document with _id = {id}. Reason: missing required fields."
            )
            return None
        embedding = self.vector_store.document_codec.decode_vector(hit)
        assert embedding is not None
        return Content(
            id=hit["_id"],
            content=hit[codec.content_field],
            embedding=embedding,
            metadata=hit["metadata"],
            # We may not have `$similarity` depending on the query.
            score=hit.get("$similarity", None),
        )



import asyncio
import json
import os

import pytest
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from pytest import Parser

from tests.embeddings import BaseEmbeddings

pytest.register_assert_rewrite("graph_retriever.testing")


@pytest.fixture(scope="session")
def event_loop():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    yield loop
    loop.close()


ALL_STORES = ["mem", "opensearch", "astra", "cassandra", "chroma"]
TESTCONTAINER_STORES = ["opensearch", "cassandra"]


def pytest_addoption(parser: Parser):
    parser.addoption(
        "--stores",
        action="append",
        metavar="STORE",
        choices=ALL_STORES + ["all"],
        help="run tests for the given store (default: 'mem')",
    )
    parser.addoption(
        "--testcontainer",
        action="append",
        metavar="STORE",
        choices=TESTCONTAINER_STORES + ["none"],
        help="which stores to run testcontainers for (default: 'all')",
    )
    parser.addoption(
        "--runextras", action="store_true", default=False, help="run tests for extras"
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "extra: mark test as requiring an `extra` package"
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runextras"):
        # --runextras given in cli: do not skip extras
        return
    skip_extras = pytest.mark.skip(reason="need --runextras option to run")
    for item in items:
        if "extra" in item.keywords:
            item.add_marker(skip_extras)


@pytest.fixture(scope="session")
def enabled_stores(request: pytest.FixtureRequest) -> set[str]:
    # TODO: Use StrEnum?
    stores = request.config.getoption("--stores")

    if stores and "all" in stores:
        return set(ALL_STORES)
    elif stores:
        return set(stores)
    else:
        return {"mem"}


@pytest.fixture(scope="session")
def testcontainers(request: pytest.FixtureRequest) -> set[str]:
    # TODO: Use StrEnum?
    testcontainers = set(request.config.getoption("--testcontainer") or [])
    if testcontainers and "none" in testcontainers:
        return set()
    elif not testcontainers:
        return set(TESTCONTAINER_STORES)
    else:
        return testcontainers


def load_animal_docs() -> list[Document]:
    documents = []

    path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "../../../data/animals.jsonl")
    )
    with open(path) as file:
        for line in file:
            data = json.loads(line.strip())
            documents.append(
                Document(
                    id=data["id"],
                    page_content=data["text"],
                    metadata=data["metadata"],
                )
            )
    return documents


@pytest.fixture(scope="session")
def animal_docs() -> list[Document]:
    return load_animal_docs()


@pytest.fixture(scope="session")
def animal_embeddings() -> Embeddings:
    # This must be imported late (after registering the rewrites)
    from graph_retriever.testing.embeddings import AnimalEmbeddings

    return BaseEmbeddings(AnimalEmbeddings())



"""Utilities for creating and analyzing a graph of documents."""

from collections.abc import Sequence

import networkx as nx
from graph_retriever import Content
from graph_retriever.edges import Edge, EdgeFunction, EdgeSpec, MetadataEdgeFunction
from langchain_core.documents import Document


def _best_communities(graph: nx.DiGraph) -> list[list[str]]:
    """
    Compute the best communities in a directed graph.

    This function iteratively applies the Girvan-Newman algorithm to partition the
    graph into communities. It continues until the modularity no longer improves.

    Parameters
    ----------
    graph :
        The directed graph to analyze.

    Returns
    -------
    :
        A list of communities, where each community is a list of node IDs.
    """
    # TODO: Also continue running until the size of communities is below
    # a specified threshold?

    best_communities = [[node] for node in graph]
    if graph.number_of_edges() == 0:
        # If there are no edges, then we can't do any beter.
        return best_communities

    # TODO: stop running if we reach a target modularity and/or number of communities?
    best_modularity = float("-inf")
    for new_communities in nx.algorithms.community.girvan_newman(graph):
        new_modularity = nx.algorithms.community.modularity(graph, new_communities)
        if new_modularity > best_modularity:
            best_modularity = new_modularity
            best_communities = new_communities
        else:
            break
    return best_communities


def create_graph(
    documents: Sequence[Document],
    *,
    edges: list[EdgeSpec] | EdgeFunction,
) -> nx.DiGraph:
    """
    Create a directed graph from a sequence of documents.

    This function constructs a directed graph where each document is a node, and
    edges are defined based on relationships in the document metadata.

    Parameters
    ----------
    documents :
        A sequence of documents to add as nodes.
    edges :
        Definitions of edges to use for creating the graph or edge function to use.

    Returns
    -------
    :
        The created directed graph with documents as nodes and metadata
        relationships as edges.

    Raises
    ------
    ValueError
        If the edges are invalid.
    """
    graph = nx.DiGraph()

    edge_function: EdgeFunction
    if isinstance(edges, list):
        edge_function = MetadataEdgeFunction(edges)
    elif callable(edges):
        edge_function = edges
    else:
        raise ValueError(f"Expected `list[EdgeSpec] | EdgeFunction` but got: {edges}")

    # First pass -- index documents based on "to_fields" so we can navigate to them.
    documents_by_incoming: dict[Edge, set[str]] = {}
    outgoing_by_id: dict[str, set[Edge]] = {}
    for document in documents:
        assert document.id is not None
        graph.add_node(document.id, doc=document)

        document_edges = edge_function(
            Content(
                id=document.id,
                content=document.page_content,
                embedding=[],
                metadata=document.metadata,
            )
        )

        for incoming in document_edges.incoming:
            documents_by_incoming.setdefault(incoming, set()).add(document.id)
        outgoing_by_id[document.id] = document_edges.outgoing

    # Second pass -- add edges for each outgoing edge.
    # print(graph.nodes)
    for id, outgoing in outgoing_by_id.items():
        linked_to = set()
        for out in outgoing:
            linked_to.update(documents_by_incoming.get(out, set()))
        for target in linked_to - {id}:
            graph.add_edge(id, target)

    return graph


def group_by_community(graph: nx.DiGraph) -> list[list[Document]]:
    """
    Group documents by community inferred from the graph's structure.

    This function partitions the graph into communities using the Girvan-Newman
    algorithm and groups documents based on their community memberships.

    Paramaters
    ----------
    graph :
        The directed graph of documents.

    Returns
    -------
    :
        A list of communities, where each community is a list of documents.
    """
    # Find communities and output documents grouped by community.
    communities = _best_communities(graph)
    return [[graph.nodes[n]["doc"] for n in community] for community in communities]



"""Provides an adapter for the InMemoryVectorStore integration."""

from collections.abc import Callable, Iterable, Sequence
from typing import Any

from langchain_core.documents import Document
from langchain_core.vectorstores.in_memory import InMemoryVectorStore
from typing_extensions import override

from langchain_graph_retriever._conversion import METADATA_EMBEDDING_KEY
from langchain_graph_retriever.adapters.langchain import LangchainAdapter

SENTINEL = object()


class InMemoryAdapter(LangchainAdapter[InMemoryVectorStore]):
    """
    Adapter for InMemoryVectorStore vector store.

    This adapter integrates the LangChain In-Memory vector store with the graph
    retriever system, enabling similarity search and document retrieval.

    Parameters
    ----------
    vector_store :
        The in-memory vector store instance.
    """

    @override
    def _get(
        self, ids: Sequence[str], filter: dict[str, Any] | None = None, **kwargs
    ) -> list[Document]:
        docs: list[Document] = []

        filter_method = self._filter_method(filter)
        for doc_id in ids:
            hit = self.vector_store.store.get(doc_id)
            if hit:
                metadata = hit["metadata"]
                metadata[METADATA_EMBEDDING_KEY] = hit["vector"]

                doc = Document(
                    id=hit["id"],
                    page_content=hit["text"],
                    metadata=metadata,
                )

                if not filter_method(doc):
                    continue

                docs.append(doc)
        return docs

    @override
    def _search(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs,
    ):
        results = self.vector_store._similarity_search_with_score_by_vector(
            embedding=embedding,
            k=k,
            filter=self._filter_method(filter),
            **kwargs,
        )
        docs = [
            Document(
                id=doc.id,
                page_content=doc.page_content,
                metadata={METADATA_EMBEDDING_KEY: doc_embedding, **doc.metadata},
            )
            for doc, _score, doc_embedding in results
        ]
        return docs

    def _equals_or_contains(
        self,
        key: str,
        value: Any,
        metadata: dict[str, Any],
    ) -> bool:
        """
        Check if a key-value pair exists or if the value is contained in the metadata.

        Parameters
        ----------
        key :
            Metadata key to look for.
        value :
            Value to check for equality or containment.
        metadata :
            Metadata dictionary to inspect.

        Returns
        -------
        :
            True if and only if `metadata[key] == value` or `metadata[key]` is a
            list containing `value`.
        """
        actual = metadata.get(key, SENTINEL)
        if actual == value:
            return True

        if (
            isinstance(actual, Iterable)
            and not isinstance(actual, str | bytes)
            and value in actual
        ):
            return True

        return False

    def _filter_method(
        self, filter_dict: dict[str, str] | None = None
    ) -> Callable[[Document], bool]:
        """
        Create a filter function based on a metadata dictionary.

        Parameters
        ----------
        filter_dict :
            Dictionary specifying the filter criteria.

        Returns
        -------
        :
            A function that determines if a document matches the filter criteria.
        """
        if filter_dict is None:
            return lambda _doc: True

        def filter(doc: Document) -> bool:
            for key, value in filter_dict.items():
                if not self._equals_or_contains(key, value, doc.metadata):
                    return False
            return True

        return filter



from collections.abc import Iterator

import pytest
from graph_retriever.adapters import Adapter
from graph_retriever.testing.adapter_tests import AdapterComplianceSuite
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_graph_retriever.transformers import ShreddingTransformer


class TestChroma(AdapterComplianceSuite):
    @pytest.fixture(scope="class")
    def adapter(
        self,
        enabled_stores: set[str],
        animal_embeddings: Embeddings,
        animal_docs: list[Document],
    ) -> Iterator[Adapter]:
        if "chroma" not in enabled_stores:
            pytest.skip("Pass --stores=chroma to test Chroma")

        from langchain_chroma.vectorstores import Chroma
        from langchain_graph_retriever.adapters.chroma import (
            ChromaAdapter,
        )

        shredder = ShreddingTransformer()
        docs = list(shredder.transform_documents(animal_docs))
        store = Chroma.from_documents(
            docs,
            animal_embeddings,
            collection_name="animals",
            # Use `cosine` metric for consistency with other systems.
            # Default was L2.
            collection_metadata={"hnsw:space": "cosine"},
        )

        yield ChromaAdapter(store, shredder, nested_metadata_fields={"keywords"})

        store.delete_collection()



import dataclasses
import os
import time
from collections.abc import Iterator

import pytest
from astrapy.authentication import StaticTokenProvider
from graph_retriever.testing.adapter_tests import (
    AdapterComplianceCase,
    AdapterComplianceSuite,
)
from langchain_astradb.utils.vector_store_codecs import _DefaultVSDocumentCodec
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_graph_retriever.adapters.astra import AstraAdapter, _QueryHelper
from typing_extensions import override

TEST_CODEC = _DefaultVSDocumentCodec("page_content", ignore_invalid_documents=True)


def test_create_ids_query_no_user() -> None:
    query_helper = _QueryHelper(TEST_CODEC, {})

    assert query_helper.create_ids_query(["1"]) == {"_id": "1"}

    query = query_helper.create_ids_query(["1", "2", "3"])
    assert query is not None
    assert query["_id"]["$in"] == ["1", "2", "3"]


def test_create_ids_query_user() -> None:
    query_helper = _QueryHelper(TEST_CODEC, {"answer": 42})

    assert query_helper.create_ids_query(["1"]) == {
        "$and": [
            {"_id": "1"},
            {"metadata.answer": 42},
        ]
    }

    query = query_helper.create_ids_query(["1", "2", "3"])
    assert query is not None
    assert query["$and"][0]["_id"]["$in"] == ["1", "2", "3"]


def test_create_metadata_query_no_user() -> None:
    query_helper = _QueryHelper(TEST_CODEC, {})

    assert query_helper.create_metadata_query({}) is None
    assert query_helper.create_metadata_query({"foo": []}) is None
    assert query_helper.create_metadata_query({"foo": [5]}) == {"metadata.foo": 5}

    query = query_helper.create_metadata_query({"foo": [5, 6]})
    assert query is not None
    assert sorted(query["metadata.foo"]["$in"]) == [5, 6]

    assert query_helper.create_metadata_query({"foo": [5], "bar": [7]}) == {
        "$or": [
            {"metadata.foo": 5},
            {"metadata.bar": 7},
        ],
    }

    query = query_helper.create_metadata_query({"foo": [5, 6], "bar": [7, 8]})
    assert query is not None
    assert sorted(query["$or"][0]["metadata.foo"]["$in"]) == [5, 6]
    assert sorted(query["$or"][1]["metadata.bar"]["$in"]) == [7, 8]

    query = query_helper.create_metadata_query({"foo": list(range(0, 200))})
    assert query is not None
    assert sorted(query["$or"][0]["metadata.foo"]["$in"]) == list(range(0, 100))
    assert sorted(query["$or"][1]["metadata.foo"]["$in"]) == list(range(100, 200))


def test_create_metadata_query_user() -> None:
    query_helper = _QueryHelper(TEST_CODEC, {"answer": 42})

    assert query_helper.create_metadata_query({}) is None
    assert query_helper.create_metadata_query({"foo": []}) is None
    assert query_helper.create_metadata_query({"foo": [5]}) == {
        "$and": [
            {"metadata.foo": 5},
            {"metadata.answer": 42},
        ],
    }

    query = query_helper.create_metadata_query({"foo": [5, 6]})
    assert query is not None
    assert query["$and"][0]["metadata.foo"]["$in"] == [5, 6]
    assert query["$and"][1] == {"metadata.answer": 42}

    assert query_helper.create_metadata_query({"foo": [5], "bar": [7]}) == {
        "$and": [
            {
                "$or": [
                    {"metadata.foo": 5},
                    {"metadata.bar": 7},
                ],
            },
            {"metadata.answer": 42},
        ],
    }

    query = query_helper.create_metadata_query({"foo": [5, 6], "bar": [7, 8]})
    assert query is not None
    assert query["$and"][0]["$or"][0]["metadata.foo"]["$in"] == [5, 6]
    assert query["$and"][0]["$or"][1]["metadata.bar"]["$in"] == [7, 8]
    assert query["$and"][1] == {"metadata.answer": 42}

    query = query_helper.create_metadata_query({"foo": list(range(0, 200))})
    assert query is not None
    assert query["$and"][0]["$or"][0]["metadata.foo"]["$in"] == list(range(0, 100))
    assert query["$and"][0]["$or"][1]["metadata.foo"]["$in"] == list(range(100, 200))
    assert query["$and"][1] == {"metadata.answer": 42}


@dataclasses.dataclass
class _AstraConfig:
    token: StaticTokenProvider
    keyspace: str
    api_endpoint: str


@pytest.fixture(scope="module")
def astra_config(enabled_stores: set[str]) -> Iterator[_AstraConfig | None]:
    if "astra" not in enabled_stores:
        pytest.skip("Pass --stores=astra to test Astra")
        return

    from astrapy import AstraDBDatabaseAdmin
    from dotenv import load_dotenv

    load_dotenv()

    token = StaticTokenProvider(os.environ["ASTRA_DB_APPLICATION_TOKEN"])
    keyspace = os.environ.get("ASTRA_DB_KEYSPACE", "default_keyspace")
    api_endpoint = os.environ["ASTRA_DB_API_ENDPOINT"]

    admin = AstraDBDatabaseAdmin(api_endpoint=api_endpoint, token=token)
    admin.create_keyspace(keyspace)

    # Sometimes the creation of the store fails because the keyspace isn't
    # created yet. To avoid that, poll the list of keyspaces until we
    # confirm it is created.
    found = False
    t_end = time.time() + 5  # run 5 seconds
    while time.time() < t_end:
        keyspaces = admin.list_keyspaces()
        if keyspace in keyspaces:
            found = True
            break

        print(f"Waiting for keyspace '{keyspace}'...")  # noqa: T201
        time.sleep(0.01)

    assert found, f"Keyspace '{keyspace}' not created"
    yield _AstraConfig(token=token, keyspace=keyspace, api_endpoint=api_endpoint)

    admin.drop_keyspace(keyspace)


class TestAstraAdapter(AdapterComplianceSuite):
    @pytest.fixture(scope="class")
    def adapter(
        self,
        animal_embeddings: Embeddings,
        animal_docs: list[Document],
        astra_config: _AstraConfig,
    ) -> Iterator["AstraAdapter"]:
        from langchain_astradb import AstraDBVectorStore

        store = AstraDBVectorStore(
            embedding=animal_embeddings,
            collection_name="animals",
            namespace=astra_config.keyspace,
            token=astra_config.token,
            api_endpoint=astra_config.api_endpoint,
            pre_delete_collection=True,
        )
        store.add_documents(animal_docs)

        yield AstraAdapter(store)

        store.delete_collection()


VECTORIZE_EXPECTATION_OVERRIDES: dict[tuple[str, str], list[str]] = {
    ("search", "basic"): ["alpaca", "cat", "chicken", "horse"],
    ("asearch", "basic"): ["alpaca", "cat", "chicken", "horse"],
    ("search_with_embedding", "basic"): ["alpaca", "cat", "chicken", "horse"],
    ("asearch_with_embedding", "basic"): ["alpaca", "cat", "chicken", "horse"],
    ("adjacent", "metadata_and_id"): [
        "cat",
        "cobra",
        "crocodile",
        "gecko",
        "iguana",
        "lizard",
    ],
    ("aadjacent", "metadata_and_id"): [
        "cat",
        "cobra",
        "crocodile",
        "gecko",
        "iguana",
        "lizard",
    ],
}


class TestAstraVectorizeAdapter(AdapterComplianceSuite):
    @override
    def expected(self, method: str, case: AdapterComplianceCase) -> list[str]:
        # Since vectorize currently requires a server-side embedding model, we
        # need to change the expectations a little to reflect the embeddings
        # that are actually computed.
        return VECTORIZE_EXPECTATION_OVERRIDES.get((method, case.id), []) or (
            super().expected(method, case)
        )

    @pytest.fixture(scope="class")
    def adapter(
        self,
        animal_docs: list[Document],
        astra_config: _AstraConfig,
    ) -> Iterator["AstraAdapter"]:
        from astrapy.info import CollectionVectorServiceOptions
        from langchain_astradb import AstraDBVectorStore

        service = CollectionVectorServiceOptions(
            provider="nvidia",
            model_name="NV-Embed-QA",
        )

        store = AstraDBVectorStore(
            collection_name="animals_vectorize",
            collection_vector_service_options=service,
            namespace=astra_config.keyspace,
            token=astra_config.token,
            api_endpoint=astra_config.api_endpoint,
            pre_delete_collection=True,
        )
        store.add_documents(animal_docs)

        yield AstraAdapter(store)

        store.delete_collection()



from graph_retriever.content import Content
from graph_retriever.types import Node
from langchain_core.documents import Document

METADATA_EMBEDDING_KEY = "__embedding"


def node_to_doc(node: Node) -> Document:
    return Document(
        id=node.id,
        page_content=node.content,
        metadata={**node.extra_metadata, **node.metadata},
    )


def doc_to_content(doc: Document, *, embedding: list[float] | None = None) -> Content:
    """
    Convert a LangChain document to a `Content`.

    Parameters
    ----------
    doc :
        The document to convert.

    embedding :
        The embedding of the document. If not provided, the
        `doc.metadata[METADATA_EMBEDDING_KEY]` should be set to the embedding.

    Returns
    -------
    :
        The converted content.
    """
    assert doc.id is not None

    if embedding is None:
        embedding = doc.metadata.pop(METADATA_EMBEDDING_KEY)
    assert embedding is not None

    return Content(
        id=doc.id,
        content=doc.page_content,
        embedding=embedding,
        metadata=doc.metadata,
    )



"""Provides an adapter for Chroma vector store integration."""

from collections.abc import Sequence
from typing import Any

from langchain_core.documents import Document
from typing_extensions import override

from langchain_graph_retriever._conversion import METADATA_EMBEDDING_KEY
from langchain_graph_retriever.adapters.langchain import ShreddedLangchainAdapter

try:
    from langchain_chroma import Chroma
except (ImportError, ModuleNotFoundError):
    msg = "please `pip install langchain-chroma`"
    raise ImportError(msg)


class ChromaAdapter(ShreddedLangchainAdapter[Chroma]):
    """
    Adapter for [Chroma](https://www.trychroma.com/) vector store.

    This adapter integrates the LangChain Chroma vector store with the
    graph retriever system, allowing for similarity search and document retrieval.

    Parameters
    ----------
    vector_store :
        The Chroma vector store instance.
    shredder: ShreddingTransformer, optional
        An instance of the ShreddingTransformer used for doc insertion.
        If not passed then a default instance of ShreddingTransformer is used.
    """

    @override
    def update_filter_hook(
        self, filter: dict[str, Any] | None
    ) -> dict[str, Any] | None:
        filter = super().update_filter_hook(filter)
        if not filter or len(filter) <= 1:
            return filter

        conjoined = [{k: v} for k, v in filter.items()]
        return {"$and": conjoined}

    @override
    def _search(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        try:
            from chromadb.api.types import IncludeEnum
        except (ImportError, ModuleNotFoundError):
            msg = "please `pip install chromadb`"
            raise ImportError(msg)

        if k > self.vector_store._collection.count():
            k = self.vector_store._collection.count()
        if k == 0:
            return []

        results = self.vector_store._collection.query(
            query_embeddings=embedding,  # type: ignore
            n_results=k,
            where=filter,  # type: ignore
            include=[
                IncludeEnum.documents,
                IncludeEnum.metadatas,
                IncludeEnum.embeddings,
            ],
            **kwargs,
        )

        docs: list[Document] = []
        # type-hint: (str, Dict[str, Any], str, ndarray)
        for content, metadata, id, emb in zip(
            results["documents"][0],  # type: ignore
            results["metadatas"][0],  # type: ignore
            results["ids"][0],  # type: ignore
            results["embeddings"][0],  # type: ignore
        ):
            docs.append(
                Document(
                    id=id,
                    page_content=content,
                    metadata={
                        METADATA_EMBEDDING_KEY: emb.tolist(),
                        **metadata,
                    },
                )
            )
        return docs

    @override
    def _get(
        self, ids: Sequence[str], filter: dict[str, Any] | None = None, **kwargs: Any
    ) -> list[Document]:
        results = self.vector_store.get(
            ids=list(ids),
            include=["embeddings", "metadatas", "documents"],
            where=self.update_filter_hook(filter),
            **kwargs,
        )
        docs = [
            Document(
                id=id,
                page_content=content,
                metadata={METADATA_EMBEDDING_KEY: emb.tolist(), **metadata},
            )
            for (content, metadata, id, emb) in zip(
                results["documents"],
                results["metadatas"],
                results["ids"],
                results["embeddings"],
            )
            # type-hint: (str, Dict[str, Any], str, ndarray)
        ]

        return docs



import pytest
from langchain_core.documents import Document


@pytest.mark.extra
def test_transform_documents(animal_docs: list[Document]):
    from langchain_graph_retriever.transformers.spacy import SpacyNERTransformer
    from spacy.language import Language  # type: ignore
    from spacy.tokens import Doc, Span  # type: ignore
    from spacy.vocab import Vocab  # type: ignore

    class FakeLanguage(Language):
        def __init__(self):
            pass

        def __call__(self, text: str | Doc, **kwargs) -> Doc:
            vocab = Vocab()
            assert isinstance(text, str)
            doc = Doc(vocab=vocab, words=text.split())
            doc.ents = [
                Span(doc, start=0, end=1, label="first"),
                Span(doc, start=1, end=2, label="second"),
                Span(doc, start=2, end=3, label="third"),
            ]
            return doc

    first_doc = animal_docs[0].model_copy()

    fake_model = FakeLanguage()

    transformer = SpacyNERTransformer(model=fake_model, metadata_key="spacey")
    transformed_docs = transformer.transform_documents(animal_docs)
    assert "spacey" in transformed_docs[0].metadata
    assert "first: the" in transformed_docs[0].metadata["spacey"]
    assert "second: aardvark" in transformed_docs[0].metadata["spacey"]
    assert "third: is" in transformed_docs[0].metadata["spacey"]

    transformer = SpacyNERTransformer(
        model=fake_model, metadata_key="spacey", include_labels=set(["first"])
    )
    transformed_docs = transformer.transform_documents(animal_docs)
    assert "spacey" in transformed_docs[0].metadata
    assert "first: the" in transformed_docs[0].metadata["spacey"]
    assert "second: aardvark" not in transformed_docs[0].metadata["spacey"]
    assert "third: is" not in transformed_docs[0].metadata["spacey"]

    transformer = SpacyNERTransformer(
        model=fake_model, metadata_key="spacey", exclude_labels=set(["first"])
    )
    transformed_docs = transformer.transform_documents(animal_docs)
    assert "spacey" in transformed_docs[0].metadata
    assert "first: the" not in transformed_docs[0].metadata["spacey"]
    assert "second: aardvark" in transformed_docs[0].metadata["spacey"]
    assert "third: is" in transformed_docs[0].metadata["spacey"]

    transformer = SpacyNERTransformer(model=fake_model, metadata_key="spacey", limit=1)
    transformed_docs = transformer.transform_documents(animal_docs)
    assert len(transformed_docs[0].metadata["spacey"]) == 1

    with pytest.raises(ValueError, match="Invalid model"):
        SpacyNERTransformer(model={})  # type: ignore

    # confirm original docs aren't modified
    assert first_doc == animal_docs[0]



"""Provides an adapter for Cassandra vector store integration."""

from collections.abc import Sequence
from typing import Any

from typing_extensions import override

try:
    from langchain_community.vectorstores.cassandra import Cassandra
except (ImportError, ModuleNotFoundError):
    raise ImportError("please `pip install langchain-community cassio`")

from langchain_core.documents import Document

from langchain_graph_retriever._conversion import METADATA_EMBEDDING_KEY
from langchain_graph_retriever.adapters.langchain import ShreddedLangchainAdapter


class CassandraAdapter(ShreddedLangchainAdapter[Cassandra]):
    """
    Adapter for the [Apache Cassandra](https://cassandra.apache.org/) vector store.

    This class integrates the LangChain Cassandra vector store with the graph
    retriever system, providing functionality for similarity search and document
    retrieval.

    Parameters
    ----------
    vector_store :
        The Cassandra vector store instance.
    shredder: ShreddingTransformer, optional
        An instance of the ShreddingTransformer used for doc insertion.
        If not passed then a default instance of ShreddingTransformer is used.
    """

    @override
    def _search(  # type: ignore
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        results = self._similarity_search_with_embedding_id_by_vector(
            embedding=embedding,
            k=k,
            filter=filter,
            **kwargs,
        )
        docs: list[Document] = []
        for doc, embedding, id in results:
            doc.metadata[METADATA_EMBEDDING_KEY] = embedding
            doc.id = id
            docs.append(doc)
        return docs

    def _similarity_search_with_embedding_id_by_vector(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        body_search: str | list[str] | None = None,
    ) -> list[tuple[Document, list[float], str]]:
        kwargs: dict[str, Any] = {}
        if filter is not None:
            kwargs["metadata"] = filter
        if body_search is not None:
            kwargs["body_search"] = body_search

        if k == 0:
            return []

        hits = self.vector_store.table.ann_search(
            vector=embedding,
            n=k,
            **kwargs,
        )
        return [
            (
                self.vector_store._row_to_document(row=hit),
                hit["vector"],
                hit["row_id"],
            )
            for hit in hits
        ]

    @override
    async def _asearch(  # type: ignore
        self, **kwargs: Any
    ) -> list[Document]:
        results = (
            await self.vector_store.asimilarity_search_with_embedding_id_by_vector(
                **kwargs
            )
        )
        docs: list[Document] = []
        for doc, embedding, id in results:
            doc.metadata[METADATA_EMBEDDING_KEY] = embedding
            doc.id = id
            docs.append(doc)
        return docs

    @override
    def _get(
        self, ids: Sequence[str], filter: dict[str, Any] | None = None, **kwargs: Any
    ) -> list[Document]:
        filter = self.update_filter_hook(filter)
        docs: list[Document] = []
        for id in ids:
            args: dict[str, Any] = {"row_id": id}
            if filter:
                args["metadata"] = filter
            row = self.vector_store.table.get(**args)
            if row is not None:
                docs.append(self._row_to_doc(row))
        return docs

    @override
    async def _aget(
        self, ids: Sequence[str], filter: dict[str, Any] | None = None, **kwargs: Any
    ) -> list[Document]:
        filter = self.update_filter_hook(filter)
        docs: list[Document] = []
        for id in ids:
            args: dict[str, Any] = {"row_id": id}
            if filter:
                args["metadata"] = filter
            row = await self.vector_store.table.aget(**args)
            if row is not None:
                docs.append(self._row_to_doc(row))
        return docs

    def _row_to_doc(self, row: Any) -> Document:
        return Document(
            id=row["row_id"],
            page_content=row["body_blob"],
            metadata={
                METADATA_EMBEDDING_KEY: row["vector"],
                **row["metadata"],
            },
        )



import pytest
from graph_retriever.adapters import Adapter
from graph_retriever.testing.adapter_tests import AdapterComplianceSuite
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores.in_memory import InMemoryVectorStore


class TestInMemory(AdapterComplianceSuite):
    @pytest.fixture(scope="class")
    def adapter(
        self,
        enabled_stores: set[str],
        animal_embeddings: Embeddings,
        animal_docs: list[Document],
    ) -> Adapter:
        if "mem" not in enabled_stores:
            pytest.skip("Pass --stores=mem to test InMemory")

        from langchain_graph_retriever.adapters.in_memory import (
            InMemoryAdapter,
        )

        store = InMemoryVectorStore.from_documents(animal_docs, animal_embeddings)
        return InMemoryAdapter(store)



from collections.abc import Iterable

import pytest
from graph_retriever.adapters import Adapter
from graph_retriever.testing.adapter_tests import AdapterComplianceSuite
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings


class TestOpenSearch(AdapterComplianceSuite):
    @pytest.fixture(scope="class")
    def adapter(
        self,
        request: pytest.FixtureRequest,
        enabled_stores: set[str],
        testcontainers: set[str],
        animal_embeddings: Embeddings,
        animal_docs: list[Document],
    ) -> Iterable[Adapter]:
        if "opensearch" not in enabled_stores:
            pytest.skip("Pass --stores=opensearch to test OpenSearch")

        from langchain_community.vectorstores import OpenSearchVectorSearch
        from langchain_graph_retriever.adapters.open_search import (
            OpenSearchAdapter,
        )

        if "opensearch" in testcontainers:
            from testcontainers.opensearch import OpenSearchContainer  # type: ignore

            # If the admin password doesn't pass the length and regex requirements
            # starting the container will hang (`docker ps <container_id>` to debug).
            container = OpenSearchContainer(
                image="opensearchproject/opensearch:2.18.0",
                initial_admin_password="SomeRandomP4ssword",
            )
            container.start()
            request.addfinalizer(lambda: container.stop())

            config = container.get_config()
            opensearch_url = f"http://{config['host']}:{config['port']}"
            kwargs = {"http_auth": (config["username"], config["password"])}
        else:
            opensearch_url = "http://localhost:9200"
            kwargs = {}

        store = OpenSearchVectorSearch(
            opensearch_url=opensearch_url,
            index_name="animals",
            embedding_function=animal_embeddings,
            engine="faiss",
            **kwargs,
        )
        store.add_documents(animal_docs)

        yield OpenSearchAdapter(store)

        if store.index_exists():
            store.delete_index()



"""Provides a graph-based retriever combining vector search and graph traversal."""

import dataclasses
from collections.abc import Sequence
from functools import cached_property
from typing import (
    Any,
)

from graph_retriever import atraverse, traverse
from graph_retriever.adapters import Adapter
from graph_retriever.edges import EdgeFunction, EdgeSpec
from graph_retriever.strategies import Eager, Strategy
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever
from langchain_core.vectorstores.base import VectorStore
from pydantic import ConfigDict, computed_field, model_validator
from typing_extensions import Self

from langchain_graph_retriever._conversion import node_to_doc
from langchain_graph_retriever.adapters.inference import infer_adapter


# this class uses pydantic, so store must be provided at init time.
class GraphRetriever(BaseRetriever):
    """
    Retriever combining vector search and graph traversal.

    The [GraphRetriever][langchain_graph_retriever.GraphRetriever] class
    retrieves documents by first performing a vector search to identify relevant
    documents, followed by graph traversal to explore their connections. It
    supports multiple traversal strategies and integrates seamlessly with
    LangChain's retriever framework.

    Attributes
    ----------
    store :
        The adapter or vector store used for document retrieval.
    edges :
        A list of [EdgeSpec][graph_retriever.edges.EdgeSpec] for use in creating a
        [MetadataEdgeFunction][graph_retriever.edges.MetadataEdgeFunction],
        or an [EdgeFunction][graph_retriever.edges.EdgeFunction].
    strategy :
        The traversal strategy to use.
    """

    store: Adapter | VectorStore
    edges: list[EdgeSpec] | EdgeFunction = []
    strategy: Strategy = Eager()

    # Capture the extra fields in `self.model_extra` rather than ignoring.
    model_config = ConfigDict(extra="allow")

    # Move the `k` extra argument (if any) to the strategy.
    @model_validator(mode="after")
    def apply_extra(self) -> Self:
        """
        Apply extra configuration to the traversal strategy.

        This method captures additional fields provided in the `model_extra` argument
        and applies them to the current traversal strategy. Any extra fields are
        cleared after they are applied.

        Returns
        -------
        :
            The updated GraphRetriever instance.
        """
        if self.model_extra:
            self.strategy = dataclasses.replace(self.strategy, **self.model_extra)
            self.model_extra.clear()
        return self

    @computed_field  # type: ignore
    @cached_property
    def adapter(self) -> Adapter:
        """The adapter to use during traversals."""
        return infer_adapter(self.store)

    def _get_relevant_documents(
        self,
        query: str,
        *,
        edges: list[EdgeSpec] | EdgeFunction | None = None,
        initial_roots: Sequence[str] = (),
        filter: dict[str, Any] | None = None,
        store_kwargs: dict[str, Any] = {},
        **kwargs: Any,
    ) -> list[Document]:
        """
        Retrieve doc nodes using graph traversal and similarity search.

        This method first retrieves documents based on similarity to the query, and
        then applies a traversal strategy to explore connected nodes in the graph.

        Notes
        -----
        You can execute this method by calling `.invoke()` on the retriever.

        Parameters
        ----------
        query :
            The query string to search for.
        edges :
            Override edge definitions for this invocation.
        initial_roots :
            Document IDs to use as initial roots. The top `adjacent_k` nodes
            connected to each root are included in the initial candidates.
        filter :
            Optional metadata filter to apply.
        store_kwargs :
            Additional keyword arguments for the store.
        kwargs :
            Additional arguments for configuring the traversal strategy.

        Returns
        -------
        :
            The retrieved documents.

        Raises
        ------
        ValueError
            If edges weren't provided in this call or the constructor.
        """
        edges = edges or self.edges
        if edges is None:
            raise ValueError("'edges' must be provided in this call or the constructor")

        nodes = traverse(
            query=query,
            edges=edges,
            strategy=Strategy.build(base_strategy=self.strategy, **kwargs),
            store=self.adapter,
            metadata_filter=filter,
            initial_root_ids=initial_roots,
            store_kwargs=store_kwargs,
        )
        return [node_to_doc(n) for n in nodes]

    async def _aget_relevant_documents(
        self,
        query: str,
        *,
        edges: list[EdgeSpec] | EdgeFunction | None = None,
        initial_roots: Sequence[str] = (),
        filter: dict[str, Any] | None = None,
        store_kwargs: dict[str, Any] = {},
        **kwargs: Any,
    ) -> list[Document]:
        """
        Asynchronously retrieve doc nodes using graph traversal and similarity search.

        This method first retrieves documents based on similarity to the query, and
        then applies a traversal strategy to explore connected nodes in the graph.

        Notes
        -----
        You can execute this method by calling `.ainvoke()` on the retriever.

        Parameters
        ----------
        query :
            The query string to search for.
        edges :
            Override edge definitions for this invocation.
        initial_roots :
            Document IDs to use as initial roots. The top `adjacent_k` nodes
            connected to each root are included in the initial candidates.
        filter :
            Optional metadata filter to apply.
        store_kwargs :
            Additional keyword arguments for the store.
        kwargs :
            Additional arguments for configuring the traversal strategy.

        Returns
        -------
        :
            The retrieved documents.

        Raises
        ------
        ValueError
            If edges weren't provided in this call or the constructor.
        """
        edges = edges or self.edges
        if edges is None:
            raise ValueError("'edges' must be provided in this call or the constructor")
        nodes = await atraverse(
            query=query,
            edges=edges,
            strategy=Strategy.build(base_strategy=self.strategy, **kwargs),
            store=self.adapter,
            metadata_filter=filter,
            initial_root_ids=initial_roots,
            store_kwargs=store_kwargs,
        )
        return [node_to_doc(n) for n in nodes]



from collections.abc import Callable

from langchain_core.embeddings import Embeddings


class BaseEmbeddings(Embeddings):
    def __init__(self, embedding: Callable[[str], list[float]]) -> None:
        self.embedding = embedding

    def embed_query(self, text: str) -> list[float]:
        return self.embedding(text)

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        return [self.embedding(text) for text in texts]



from __future__ import annotations

from typing import Any

import langchain_astradb
import pytest
from graph_retriever.adapters import Adapter
from langchain_chroma import Chroma
from langchain_community.vectorstores.cassandra import Cassandra
from langchain_community.vectorstores.opensearch_vector_search import (
    OpenSearchVectorSearch,
)
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings, FakeEmbeddings
from langchain_core.vectorstores.in_memory import InMemoryVectorStore, VectorStore
from langchain_graph_retriever.adapters.astra import AstraAdapter
from langchain_graph_retriever.adapters.cassandra import CassandraAdapter
from langchain_graph_retriever.adapters.chroma import ChromaAdapter
from langchain_graph_retriever.adapters.inference import (
    _infer_adapter_name,
    infer_adapter,
)
from langchain_graph_retriever.adapters.open_search import OpenSearchAdapter
from typing_extensions import override


def test_infer_in_memory():
    store = InMemoryVectorStore(FakeEmbeddings(size=4))
    adapter = infer_adapter(store)
    assert isinstance(adapter, Adapter)


@pytest.mark.parametrize(
    "cls,adapter_cls",
    [
        (langchain_astradb.AstraDBVectorStore, AstraAdapter),
        (Cassandra, CassandraAdapter),
        (Chroma, ChromaAdapter),
        (OpenSearchVectorSearch, OpenSearchAdapter),
    ],
)
def test_infer_adapter_name(cls: type, adapter_cls: type) -> None:
    module_name, class_name = _infer_adapter_name(cls)
    assert module_name == adapter_cls.__module__
    assert class_name == adapter_cls.__name__


class UnsupportedVectorStore(VectorStore):
    @classmethod
    @override
    def from_texts(
        cls,
        texts: list[str],
        embedding: Embeddings,
        metadatas: list[dict[Any, Any]] | None = None,
        *,
        ids: list[str] | None = None,
        **kwargs: Any,
    ) -> UnsupportedVectorStore:
        return UnsupportedVectorStore()

    @override
    def similarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[Document]:
        return []


def test_infer_store_unknown():
    EXPECTED = (
        "Expected adapter or supported vector store, but got"
        f" {__name__}.UnsupportedVectorStore"
    )
    with pytest.raises(ValueError, match=EXPECTED):
        infer_adapter(UnsupportedVectorStore())



import json

from langchain_core.documents import Document
from langchain_graph_retriever.transformers.shredding import (
    DEFAULT_PATH_DELIMITER,
    DEFAULT_STATIC_VALUE,
    SHREDDED_KEYS_KEY,
    ShreddingTransformer,
)


def test_transform_documents(animal_docs: list[Document]):
    first_doc = animal_docs[0].model_copy()
    original_docs = [animal_docs[0]]
    transformer = ShreddingTransformer()

    # confirm "keywords" contains a list value in original document
    list_key = "keywords"
    original_metadata = original_docs[0].metadata
    assert isinstance(original_metadata[list_key], list)

    # pull the first keyword in that list
    first_keyword = original_metadata[list_key][0]

    # transform the document
    shredded_docs = transformer.transform_documents(original_docs)
    shredded_metadata = shredded_docs[0].metadata

    # confirm "keywords" no longer exists as a metadata key
    assert list_key not in shredded_metadata

    # confirm "keywords" exists as a shredded key
    assert list_key in json.loads(shredded_metadata[SHREDDED_KEYS_KEY])

    # confirm the shredded key has the expected value
    shredded_key = f"{list_key}{DEFAULT_PATH_DELIMITER}{first_keyword}"
    assert shredded_key in shredded_metadata
    assert shredded_metadata[shredded_key] == DEFAULT_STATIC_VALUE

    # confirm original docs aren't modified
    assert first_doc == animal_docs[0]


def test_restore_documents(animal_docs: list[Document]):
    first_doc = animal_docs[0].model_copy()
    original_docs = [animal_docs[0]]
    transformer = ShreddingTransformer()

    shredded_docs = transformer.transform_documents(original_docs)

    restored_docs = transformer.restore_documents(shredded_docs)

    assert original_docs[0] == restored_docs[0]

    # confirm original docs aren't modified
    assert first_doc == animal_docs[0]



from typing import Any

import pytest
from langchain_core.documents import Document


@pytest.mark.extra
def test_transform_documents(animal_docs: list[Document]):
    from keybert import KeyBERT  # type: ignore
    from langchain_graph_retriever.transformers.keybert import KeyBERTTransformer

    class FakeKeyBERT(KeyBERT):
        def __init__(self):
            pass

        def extract_keywords(
            self, docs: list[str], **kwargs: Any
        ) -> list[list[tuple[str, float]]]:
            return [
                [(word, len(word)) for word in set(doc.split()) if len(word) > 5]
                for doc in docs
            ]

    first_doc = animal_docs[0].model_copy()

    fake_model = FakeKeyBERT()
    transformer = KeyBERTTransformer(model=fake_model, metadata_key="keybert")

    transformed_docs = transformer.transform_documents(animal_docs)
    assert "keybert" in transformed_docs[0].metadata

    with pytest.raises(ValueError, match="Invalid model"):
        KeyBERTTransformer(model={})

    # confirm original docs aren't modified
    assert first_doc == animal_docs[0]



import pytest
from langchain_core.documents import Document
from langchain_graph_retriever.transformers.parent import ParentTransformer


def test_transform_documents():
    root = Document(id="root", page_content="test", metadata={"_path": "root"})
    h1 = Document(id="h1", page_content="test", metadata={"_path": "root.h1"})
    h1a = Document(id="h1a", page_content="test", metadata={"_path": "root.h1.a"})

    original_h1 = h1.model_copy()

    transformer = ParentTransformer(
        path_metadata_key="_path", parent_metadata_key="_parent", path_delimiter="."
    )

    transformed_root = transformer.transform_documents([root])[0]
    assert "_parent" not in transformed_root.metadata

    transformed_h1 = transformer.transform_documents([h1])[0]
    assert transformed_h1.metadata["_parent"] == "root"

    transformed_h1a = transformer.transform_documents([h1a])[0]
    assert transformed_h1a.metadata["_parent"] == "root.h1"

    transformer = ParentTransformer()
    with pytest.raises(ValueError, match="path not found in metadata"):
        transformer.transform_documents([root])

    # confirm original docs aren't modified
    assert original_h1 == h1



from typing import cast

import networkx as nx
import pytest
from graph_retriever.edges import Id
from langchain_core.documents import Document
from langchain_graph_retriever.document_graph import create_graph, group_by_community


@pytest.fixture
def docs() -> list[Document]:
    doc0 = Document(
        id="doc0",
        page_content="Typical Greetings",
        metadata={
            "incoming": "parent",
        },
    )

    doc1 = Document(
        id="doc1",
        page_content="Hello World",
        metadata={
            "outgoing": "parent",
            "keywords": ["greeting", "world"],
            "mentions": ["doc0"],
        },
    )

    doc2 = Document(
        id="doc2",
        page_content="Hello Earth",
        metadata={
            "outgoing": "parent",
            "keywords": ["greeting", "earth"],
            "mentions": ["doc1", "doc0"],
        },
    )
    return [doc0, doc1, doc2]


def test_create_graph(docs: list[Document]):
    graph = create_graph(
        docs,
        edges=[("outgoing", "incoming")],
    )
    assert sorted(nx.to_edgelist(graph)) == [
        ("doc1", "doc0", {}),
        ("doc2", "doc0", {}),
    ]

    graph = create_graph(
        docs,
        edges=[("keywords", "keywords")],
    )
    assert sorted(nx.to_edgelist(graph)) == [
        ("doc1", "doc2", {}),
        ("doc2", "doc1", {}),
    ]

    graph = create_graph(
        docs,
        edges=[("outgoing", "incoming"), ("keywords", "keywords")],
    )
    assert sorted(nx.to_edgelist(graph)) == [
        ("doc1", "doc0", {}),
        ("doc1", "doc2", {}),
        ("doc2", "doc0", {}),
        ("doc2", "doc1", {}),
    ]
    assert graph.nodes["doc1"]["doc"] == docs[1]

    graph = create_graph(
        docs,
        edges=[("mentions", Id())],
    )
    assert sorted(nx.to_edgelist(graph)) == [
        ("doc1", "doc0", {}),
        ("doc2", "doc0", {}),
        ("doc2", "doc1", {}),
    ]


def test_communities(animal_docs: list[Document]):
    graph = create_graph(
        animal_docs,
        edges=[("habitat", "habitat")],
    )
    communities = group_by_community(graph)

    community_ids = [sorted([cast(str, d.id) for d in c]) for c in communities]
    assert community_ids == [
        ["aardvark"],
        ["albatross", "barracuda", "crab"],
        ["alligator"],
        ["alpaca"],
        ["ant"],
        ["anteater"],
        ["antelope", "buffalo", "coyote", "hedgehog"],
        ["armadillo"],
        ["baboon"],
        ["badger"],
        ["bat"],
        ["bear"],
        ["beaver"],
        ["bee"],
        ["beetle"],
        ["bison"],
        ["blue jay"],
        ["boar"],
        ["bobcat", "cobra", "deer", "elk", "mongoose"],
        ["butterfly"],
        ["camel"],
        [
            "capybara",
            "crane",
            "crocodile",
            "dragonfly",
            "duck",
            "frog",
            "heron",
            "newt",
        ],
        ["caribou"],
        ["cassowary", "jaguar"],
        ["cat"],
        ["caterpillar"],
        ["chameleon", "chimpanzee", "gecko", "gorilla", "iguana", "leopard"],
        ["cheetah", "gazelle", "hyena", "lion", "ostrich"],
        ["chicken"],
        ["chinchilla"],
        ["cockroach"],
        ["crow"],
        ["dingo"],
        ["dog"],
        ["dolphin", "jellyfish"],
        ["donkey"],
        ["dove"],
        ["eagle"],
        ["eel"],
        ["elephant"],
        ["emu"],
        ["falcon"],
        ["ferret"],
        ["finch"],
        ["fish"],
        ["flamingo"],
        ["fox"],
        ["giraffe"],
        ["goat"],
        ["goose"],
        ["grasshopper"],
        ["guinea pig"],
        ["hamster"],
        ["hawk"],
        ["hippopotamus"],
        ["hornet"],
        ["horse"],
        ["hummingbird"],
        ["jackal"],
        ["kangaroo"],
        ["koala"],
        ["komodo dragon"],
        ["lark"],
        ["lemur"],
        ["lizard"],
        ["lobster"],
        ["magpie"],
        ["manatee"],
        ["moose"],
        ["mosquito"],
        ["narwhal"],
        ["octopus"],
    ]


def test_communities_disconnected_graph():
    graph = nx.DiGraph()
    graph.add_node("n1", doc=Document(id="n1", page_content="n1"))
    graph.add_node("n2", doc=Document(id="n2", page_content="n2"))

    communities = group_by_community(graph)

    community_ids = [sorted([cast(str, d.id) for d in c]) for c in communities]
    assert community_ids == [["n1"], ["n2"]]



"""Shredding Transformer for sequence-based metadata fields."""

import json
from collections.abc import Sequence
from typing import Any

from langchain_core.documents import BaseDocumentTransformer, Document
from typing_extensions import override

SHREDDED_KEYS_KEY = "__shredded_keys"
DEFAULT_PATH_DELIMITER = "→"  # unicode 2192
DEFAULT_STATIC_VALUE = "§"  # unicode 00A7


class ShreddingTransformer(BaseDocumentTransformer):
    """
    Shreds sequence-based metadata fields.

    Certain vector stores do not support storing or searching on metadata fields
    with sequence-based values. This transformer converts sequence-based fields
    into simple metadata values.

    Example
    -------
    An example of how to use this transformer exists
    [HERE](../../guide/transformers.md#shreddingtransformer) in the guide.

    Parameters
    ----------
    keys :
        A set of metadata keys to shred.
        If empty, all sequence-based fields will be shredded.
    path_delimiter :
        The path delimiter to use when building shredded keys.
    static_value :
        The value to set on each shredded key.
    """

    def __init__(
        self,
        *,
        keys: set[str] = set(),
        path_delimiter: str = DEFAULT_PATH_DELIMITER,
        static_value: Any = DEFAULT_STATIC_VALUE,
    ):
        self.keys = keys
        self.path_delimiter = path_delimiter
        self.static_value = static_value

    @override
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        transformed_docs = []
        for document in documents:
            new_doc = Document(id=document.id, page_content=document.page_content)
            shredded_keys: list[str] = []
            for key, value in document.metadata.items():
                is_nested_sequence = isinstance(value, Sequence) and not isinstance(
                    value, str | bytes
                )
                should_shred = (not self.keys) or (key in self.keys)
                if is_nested_sequence and should_shred:
                    shredded_keys.append(key)
                    for item in value:
                        new_doc.metadata[self.shredded_key(key=key, value=item)] = (
                            self.shredded_value()
                        )
                else:
                    new_doc.metadata[key] = value
            if len(shredded_keys) > 0:
                new_doc.metadata[SHREDDED_KEYS_KEY] = json.dumps(shredded_keys)
            transformed_docs.append(new_doc)

        return transformed_docs

    def restore_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        """
        Restore documents transformed by the ShreddingTransformer.

        Restore documents transformed by the ShreddingTransformer back to
        their original state before shredding.

        Note that any non-string values inside lists will be converted to strings
        after restoring.

        Args:
            documents: A sequence of Documents to be transformed.

        Returns
        -------
        Sequence[Document]
            A sequence of transformed Documents.
        """
        restored_docs = []
        for document in documents:
            new_doc = Document(id=document.id, page_content=document.page_content)
            shredded_keys = set(
                json.loads(document.metadata.pop(SHREDDED_KEYS_KEY, "[]"))
            )

            for key, value in document.metadata.items():
                # Check if the key belongs to a shredded group
                split_key = key.split(self.path_delimiter, 1)
                if (
                    len(split_key) == 2
                    and split_key[0] in shredded_keys
                    and value == self.static_value
                ):
                    original_key, original_value = split_key
                    if original_key not in new_doc.metadata:
                        new_doc.metadata[original_key] = []
                    new_doc.metadata[original_key].append(original_value)
                else:
                    # Retain non-shredded metadata as is
                    new_doc.metadata[key] = value

            restored_docs.append(new_doc)

        return restored_docs

    def shredded_key(self, key: str, value: Any) -> str:
        """
        Get the shredded key for a key/value pair.

        Parameters
        ----------
        key :
            The metadata key to shred
        value :
            The metadata value to shred

        Returns
        -------
        str
            the shredded key
        """
        return f"{key}{self.path_delimiter}{value}"

    def shredded_value(self) -> str:
        """
        Get the shredded value for a key/value pair.

        Returns
        -------
        str
            the shredded value
        """
        return self.static_value



from collections.abc import Sequence
from typing import Any

from langchain_core.documents import BaseDocumentTransformer, Document
from typing_extensions import override


class ParentTransformer(BaseDocumentTransformer):
    """
    Adds the hierarchal Parent path to the document metadata.

    Parameters
    ----------
    path_metadata_key :
        Metadata key containing the path.
        This may correspond to paths in a file system, hierarchy in a document, etc.
    parent_metadata_key:
        Metadata key for the added parent path
    path_delimiter :
        Delimiter of items in the path.

    Example
    -------
    An example of how to use this transformer exists
    [HERE](../../guide/transformers.md#parenttransformer) in the guide.

    Notes
    -----
    Expects each document to contain its _path_ in its metadata.
    """

    def __init__(
        self,
        *,
        path_metadata_key: str = "path",
        parent_metadata_key: str = "parent",
        path_delimiter: str = "\\",
    ):
        self._path_metadata_key = path_metadata_key
        self._parent_metadata_key = parent_metadata_key
        self._path_delimiter = path_delimiter

    @override
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        results: list[Document] = []
        for document in documents:
            if self._path_metadata_key not in document.metadata:
                msg = (
                    f"path not found in metadata at {self._path_metadata_key}",
                    f" for document id: {document.id}",
                )
                raise ValueError(msg)

            path: str = document.metadata[self._path_metadata_key]
            path_parts = path.split(self._path_delimiter)
            result = document.model_copy()
            if len(path_parts) > 1:
                parent_path = self._path_delimiter.join(path_parts[0:-1])
                result.metadata[self._parent_metadata_key] = parent_path
            results.append(result)
        return results



import pytest
from langchain_core.documents import Document

test_html = """
<!DOCTYPE html>
<html><head><title>Animals of the World</title></head>
<body>
    <h1>Animals of the World</h1>
    <p>Explore the fascinating world of animals.</p>

    <h2>Mammals</h2>
    <p>The <a href="https://example.com/lion">lion</a> is the king of the jungle.</p>
    <p>The <a href="https://example.com/elephant">elephant</a> is a large animal.</p>

    <h2>Birds</h2>
    <p>The <a href="https://example.com/eagle">eagle</a> soars high in the sky.</p>
    <p>The <a href="https://example.com/penguin">penguin</a> thrives in icy areas.</p>

    <h2>Reptiles</h2>
    <p>The <a href="https://example.com/crocodile">crocodile</a> is a predator.</p>
    <p>The <a href="https://example.com/gecko">gecko</a> can climb walls.</p>
</body></html>
"""


@pytest.mark.extra
def test_transform_documents():
    from langchain_graph_retriever.transformers.html import (
        HyperlinkTransformer,
    )

    doc = Document(
        id="animal_html",
        page_content=test_html,
        metadata={"_url": "https://example.com/animals"},
    )

    original_doc = doc.model_copy()

    transformer = HyperlinkTransformer(
        url_metadata_key="_url",
        metadata_key="_hyperlinks",
    )

    transformed_doc = transformer.transform_documents([doc])[0]
    assert "_hyperlinks" in transformed_doc.metadata
    assert "https://example.com/gecko" in transformed_doc.metadata["_hyperlinks"]
    assert len(transformed_doc.metadata["_hyperlinks"]) == 6

    transformer = HyperlinkTransformer()
    with pytest.raises(ValueError, match="html document url not found in metadata"):
        transformer.transform_documents([doc])

    # confirm original docs aren't modified
    assert original_doc == doc



import typing
from collections.abc import Iterator
from typing import Union

import pytest
from graph_retriever.adapters import Adapter
from graph_retriever.testing.adapter_tests import AdapterComplianceSuite
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_graph_retriever.transformers import ShreddingTransformer

if typing.TYPE_CHECKING:
    from cassandra.cluster import Cluster  # type: ignore


@pytest.fixture(scope="module")
def cluster(
    request: pytest.FixtureRequest, enabled_stores: set[str], testcontainers: set[str]
) -> Iterator[Union["Cluster", None]]:
    if "cassandra" not in enabled_stores:
        pytest.skip("Pass --stores=cassandra to test Cassandra")
        return

    import os

    from cassandra.cluster import Cluster  # type: ignore

    if "cassandra" in testcontainers:
        from testcontainers.cassandra import CassandraContainer  # type: ignore

        container = CassandraContainer(image="cassandra:5.0.2")
        container.start()

        request.addfinalizer(lambda: container.stop())
        contact_points = container.get_contact_points()
    elif "CASSANDRA_CONTACT_POINTS" in os.environ:
        contact_points = [
            cp.strip()
            for cp in os.environ["CASSANDRA_CONTACT_POINTS"].split(",")
            if cp.strip()
        ]
    else:
        contact_points = None

    cluster = Cluster(contact_points)
    yield cluster
    cluster.shutdown()


class TestCassandraAdapter(AdapterComplianceSuite):
    @pytest.fixture(scope="class")
    def adapter(
        self,
        cluster: "Cluster",
        animal_embeddings: Embeddings,
        animal_docs: list[Document],
    ) -> Iterator[Adapter]:
        from langchain_community.vectorstores.cassandra import Cassandra
        from langchain_graph_retriever.adapters.cassandra import (
            CassandraAdapter,
        )

        session = cluster.connect()

        KEYSPACE = "graph_test_keyspace"
        session.execute(
            f"CREATE KEYSPACE IF NOT EXISTS {KEYSPACE}"
            " WITH replication = "
            "{'class': 'SimpleStrategy', 'replication_factor': 1}"
        )

        shredder = ShreddingTransformer()
        session = cluster.connect()
        session.execute(f"DROP TABLE IF EXISTS {KEYSPACE}.animals")
        store = Cassandra(
            embedding=animal_embeddings,
            session=session,
            keyspace=KEYSPACE,
            table_name="animals",
        )
        docs = list(shredder.transform_documents(animal_docs))
        store.add_documents(docs)
        yield CassandraAdapter(store, shredder, {"keywords"})

        if session:
            session.shutdown()



from typing import Any

import pytest
from langchain_core.documents import Document


@pytest.mark.extra
def test_transform_documents(animal_docs: list[Document]):
    from gliner import GLiNER  # type: ignore
    from langchain_graph_retriever.transformers.gliner import GLiNERTransformer

    class FakeGLiNER(GLiNER):
        def __init__(self):
            pass

        def batch_predict_entities(
            self, texts: list[str], **kwargs: Any
        ) -> list[list[dict[str, str]]]:
            return [[{"text": text.split()[0], "label": "first"}] for text in texts]

    first_doc = animal_docs[0].model_copy()

    fake_model = FakeGLiNER()
    transformer = GLiNERTransformer(
        ["first"], model=fake_model, metadata_key_prefix="prefix_"
    )

    transformed_docs = transformer.transform_documents(animal_docs)
    assert "prefix_first" in transformed_docs[0].metadata

    with pytest.raises(ValueError, match="Invalid model"):
        GLiNERTransformer([], model={})

    # confirm original docs aren't modified
    assert first_doc == animal_docs[0]



from __future__ import annotations

from collections.abc import Sequence
from typing import Any
from urllib.parse import urldefrag, urljoin, urlparse

from bs4 import BeautifulSoup
from bs4.element import Tag
from langchain_core.documents import BaseDocumentTransformer, Document
from typing_extensions import override


class HyperlinkTransformer(BaseDocumentTransformer):
    """
    Extracts hyperlinks from HTML content and stores them in document metadata.

    Prerequisites
    -------------

    This transformer requires the `html` extra to be installed.

    ```
    pip install -qU langchain_graph_retriever[html]
    ```

    Example
    -------
    An example of how to use this transformer exists
    [HERE](../../guide/transformers.md#hyperlinktransformer) in the guide.

    Parameters
    ----------
    url_metadata_key :
        The metadata field containing the URL of the document. Must be set
        before transforming. Needed to resolve relative paths.
    metadata_key :
        The metadata field to populate with documents linked from this content.
    drop_fragments :
        Whether fragments in URLs and links should be dropped.

    Notes
    -----
    Expects each document to contain its _URL_ in its metadata.

    """  # noqa: E501

    def __init__(
        self,
        *,
        url_metadata_key: str = "url",
        metadata_key: str = "hyperlink",
        drop_fragments: bool = True,
    ):
        self._url_metadata_key = url_metadata_key
        self._metadata_key = metadata_key
        self._drop_fragments = drop_fragments

    @staticmethod
    def _parse_url(link: Tag, page_url: str, drop_fragments: bool = True) -> str | None:
        href = link.get("href")
        if href is None:
            return None
        if isinstance(href, list) and len(href) == 1:
            href = href[0]
        if not isinstance(href, str):
            return None

        url = urlparse(href)
        if url.scheme not in ["http", "https", ""]:
            return None

        # Join the HREF with the page_url to convert relative paths to absolute.
        joined_url = str(urljoin(page_url, href))

        # Fragments would be useful if we chunked a page based on section.
        # Then, each chunk would have a different URL based on the fragment.
        # Since we aren't doing that yet, they just "break" links. So, drop
        # the fragment.
        if drop_fragments:
            return urldefrag(joined_url).url
        return joined_url

    @staticmethod
    def _parse_urls(
        soup: BeautifulSoup, page_url: str, drop_fragments: bool = True
    ) -> list[str]:
        soup_links: list[Tag] = soup.find_all("a")
        urls: set[str] = set()

        for link in soup_links:
            parsed_url = HyperlinkTransformer._parse_url(
                link, page_url=page_url, drop_fragments=drop_fragments
            )
            # Remove self links and entries for any 'a' tag that failed to parse
            # (didn't have href, or invalid domain, etc.)
            if parsed_url and parsed_url != page_url:
                urls.add(parsed_url)

        return list(urls)

    @override
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        results: list[Document] = []
        for document in documents:
            if self._url_metadata_key not in document.metadata:
                msg = (
                    f"html document url not found in metadata at "
                    f"{self._url_metadata_key} for document id: {document.id}"
                )
                raise ValueError(msg)

            page_url = document.metadata[self._url_metadata_key]
            if self._drop_fragments:
                page_url = urldefrag(page_url).url

            soup = BeautifulSoup(document.page_content, "html.parser", **kwargs)
            parsed_urls = self._parse_urls(
                soup=soup, page_url=page_url, drop_fragments=self._drop_fragments
            )

            results.append(
                Document(
                    id=document.id,
                    page_content=document.page_content,
                    metadata={self._metadata_key: parsed_urls, **document.metadata},
                )
            )
        return results



from collections.abc import Sequence
from typing import Any

import spacy  # type: ignore
from langchain_core.documents import BaseDocumentTransformer, Document
from spacy.language import Language  # type: ignore
from typing_extensions import override


class SpacyNERTransformer(BaseDocumentTransformer):
    """
    Add metadata to documents about named entities using **spaCy**.

    Identifies and labels named entities in text, extracting structured
    metadata such as organizations, locations, dates, and other key entity types.

    [**spaCy**](https://spacy.io/) is a library for Natural Language Processing
    in Python. Here it is used for Named Entity Recognition (NER) to extract values
    from document content.

    Prerequisites
    -------------

    This transformer requires the `spacy` extra to be installed.

    ```
    pip install -qU langchain_graph_retriever[spacy]
    ```

    Example
    -------
    An example of how to use this transformer exists
    [HERE](../../guide/transformers.md#spacynertransformer) in the guide.

    Parameters
    ----------
    include_labels :
        Set of entity labels to include. Will include all labels if empty.
    exclude_labels :
        Set of entity labels to exclude. Will not exclude anything if empty.
    metadata_key :
        The metadata key to store the extracted entities in.
    model :
        The spaCy model to use. Pass the name of a model to load
        or pass an instantiated spaCy model instance.

    Notes
    -----
     See spaCy docs for the selected model to determine what NER labels will be
     used. The default model
    [en_core_web_sm](https://spacy.io/models/en#en_core_web_sm-labels) produces:
    CARDINAL, DATE, EVENT, FAC, GPE, LANGUAGE, LAW, LOC, MONEY, NORP, ORDINAL,
    ORG, PERCENT, PERSON, PRODUCT, QUANTITY, TIME, WORK_OF_ART.

    """

    def __init__(
        self,
        *,
        include_labels: set[str] = set(),
        exclude_labels: set[str] = set(),
        limit: int | None = None,
        metadata_key: str = "entities",
        model: str | Language = "en_core_web_sm",
    ):
        self.include_labels = include_labels
        self.exclude_labels = exclude_labels
        self.limit = limit
        self.metadata_key = metadata_key

        if isinstance(model, str):
            if not spacy.util.is_package(model):
                spacy.cli.download(model)  # type: ignore
            self.model = spacy.load(model)
        elif isinstance(model, Language):
            self.model = model
        else:
            raise ValueError(f"Invalid model: {model}")

    @override
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        docs = []
        for doc in documents:
            results = self.model(doc.page_content).ents
            # Filter and de-duplicate entities.
            entities = list(
                {
                    f"{e.label_}: {e.text}"
                    for e in results
                    if not self.include_labels or e.label_ in self.include_labels
                    if not self.exclude_labels or e.label_ not in self.exclude_labels
                }
            )
            # Limit it, if necessary.
            if self.limit:
                entities = entities[: self.limit]
            docs.append(
                Document(
                    id=doc.id,
                    page_content=doc.page_content,
                    metadata={self.metadata_key: entities, **doc.metadata},
                )
            )
        return docs



from collections.abc import Sequence
from typing import Any

from gliner import GLiNER  # type: ignore
from langchain_core.documents import BaseDocumentTransformer, Document
from typing_extensions import override


class GLiNERTransformer(BaseDocumentTransformer):
    """
    Add metadata to documents about named entities using **GLiNER**.

    Extracts structured entity labels from text, identifying key attributes and
    categories to enrich document metadata with semantic information.

    [**GLiNER**](https://github.com/urchade/GLiNER) is a Named Entity
    Recognition (NER) model capable of identifying any entity type using a
    bidirectional transformer encoder (BERT-like).

    Prerequisites
    -------------

    This transformer requires the `gliner` extra to be installed.

    ```
    pip install -qU langchain_graph_retriever[gliner]
    ```

    Example
    -------
    An example of how to use this transformer exists
    [HERE](../../guide/transformers.md#glinertransformer) in the guide.


    Parameters
    ----------
    labels :
        List of entity kinds to extract.
    batch_size :
        The number of documents to process in each batch.
    metadata_key_prefix :
        A prefix to add to metadata keys outputted by the extractor.
        This will be prepended to the label, with the value (or values) holding the
        generated keywords for that entity kind.
    model :
        The GLiNER model to use. Pass the name of a model to load or
        pass an instantiated GLiNER model instance.

    """  # noqa: E501

    def __init__(
        self,
        labels: list[str],
        *,
        batch_size: int = 8,
        metadata_key_prefix: str = "",
        model: str | GLiNER = "urchade/gliner_mediumv2.1",
    ):
        if isinstance(model, GLiNER):
            self._model = model
        elif isinstance(model, str):
            self._model = GLiNER.from_pretrained(model)
        else:
            raise ValueError(f"Invalid model: {model}")

        self._batch_size = batch_size
        self._labels = labels
        self.metadata_key_prefix = metadata_key_prefix

    @override
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        results: list[Document] = []
        for i in range(0, len(documents), self._batch_size):
            batch = documents[i : i + self._batch_size]
            texts = [item.page_content for item in batch]
            extracted = self._model.batch_predict_entities(
                texts=texts, labels=self._labels, **kwargs
            )
            for j, entities in enumerate(extracted):
                new_metadata: dict[str, Any] = {}
                for entity in entities:
                    label = self.metadata_key_prefix + entity["label"]
                    new_metadata.setdefault(label, set()).add(entity["text"].lower())

                result = Document(
                    id=batch[j].id,
                    page_content=batch[j].page_content,
                    metadata=batch[j].metadata.copy(),
                )
                for key in new_metadata.keys():
                    result.metadata[key] = list(new_metadata[key])

                results.append(result)
        return results



"""
Package containing useful Document Transformers.

Many of these add metadata that could be useful for linking content, such as
extracting named entities or keywords from the page content.

Also includes a transformer for shredding metadata, for use with stores
that do not support querying on elements of lists.
"""

from .parent import ParentTransformer
from .shredding import ShreddingTransformer

__all__ = [
    "ShreddingTransformer",
    "ParentTransformer",
]



from collections.abc import Sequence
from typing import Any

from keybert import KeyBERT  # type: ignore
from langchain_core.documents import BaseDocumentTransformer, Document
from typing_extensions import override


class KeyBERTTransformer(BaseDocumentTransformer):
    """
    Add metadata to documents about keywords using **KeyBERT**.

    Extracts key topics and concepts from text, generating metadata that highlights
    the most relevant terms to describe the content.

    [**KeyBERT**](https://maartengr.github.io/KeyBERT) is a minimal and easy-to-use
    keyword extraction technique that leverages BERT embeddings to create keywords and
    keyphrases that are most similar to a document.

    Prerequisites
    -------------

    This transformer requires the `keybert` extra to be installed.

    ```
    pip install -qU langchain_graph_retriever[keybert]
    ```

    Example
    -------
    An example of how to use this transformer exists
    [HERE](../../guide/transformers.md#keyberttransformer) in the guide.

    Parameters
    ----------
    batch_size
        The number of documents to process in each batch.
    metadata_key :
        The name of the key used in the metadata output.
    model :
        The KeyBERT model to use. Pass the name of a model to load
        or pass an instantiated KeyBERT model instance.
    """

    def __init__(
        self,
        *,
        batch_size: int = 8,
        metadata_key: str = "keywords",
        model: str | KeyBERT = "all-MiniLM-L6-v2",
    ):
        if isinstance(model, KeyBERT):
            self._kw_model = model
        elif isinstance(model, str):
            self._kw_model = KeyBERT(model=model)
        else:
            raise ValueError(f"Invalid model: {model}")
        self._batch_size = batch_size
        self._metadata_key = metadata_key

    def _extract_keywords(
        self, docs: list[str], **kwargs
    ) -> list[list[tuple[str, float]]]:
        """Wrap the function to always return a list of responses."""
        extracted = self._kw_model.extract_keywords(docs=docs, **kwargs)
        if len(docs) == 1:
            # Even if we pass a list, if it contains one item, keybert will flatten it.
            return [extracted]
        else:
            return extracted

    @override
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        results: list[Document] = []
        for i in range(0, len(documents), self._batch_size):
            batch = documents[i : i + self._batch_size]
            texts = [item.page_content for item in batch]
            extracted = self._extract_keywords(docs=texts, **kwargs)
            for j, keywords in enumerate(extracted):
                results.append(
                    Document(
                        id=batch[j].id,
                        page_content=batch[j].page_content,
                        metadata={
                            self._metadata_key: [kw[0] for kw in keywords],
                            **batch[j].metadata,
                        },
                    )
                )
        return results



import copy
from collections.abc import Iterable

from graph_retriever.edges.metadata import Id
from graph_retriever.strategies import Mmr
from langchain_core.documents import Document
from langchain_core.embeddings import FakeEmbeddings
from langchain_core.vectorstores.in_memory import InMemoryVectorStore
from langchain_graph_retriever import GraphRetriever
from langchain_graph_retriever.adapters.in_memory import InMemoryAdapter


def test_mmr_parameters() -> None:
    # Makes sure that copying the MMR strategy creates new embeddings.
    mmr1 = Mmr()
    mmr1._query_embedding = [0.25, 0.5, 0.75]
    assert id(mmr1._nd_query_embedding) == id(mmr1._nd_query_embedding)

    mmr2 = copy.deepcopy(mmr1)
    assert id(mmr1._nd_query_embedding) != id(mmr2._nd_query_embedding)


def test_init_parameters_override_strategy() -> None:
    store = InMemoryAdapter(vector_store=InMemoryVectorStore(FakeEmbeddings(size=8)))
    retriever = GraphRetriever(store=store, edges=[], k=87)  # type: ignore[call-arg]
    assert retriever.strategy.k == 87


def test_infers_adapter() -> None:
    # Some vector stores require at least one document to be created.
    doc = Document(
        id="doc",
        page_content="lorem ipsum and whatnot",
    )
    store = InMemoryVectorStore.from_documents([doc], FakeEmbeddings(size=8))

    retriever = GraphRetriever(
        store=store,
        edges=[],
    )

    assert isinstance(retriever.adapter, InMemoryAdapter)


def sorted_ids(docs: Iterable[Document]) -> list[str]:
    ids = []
    for doc in docs:
        assert doc.id is not None
        ids.append(doc.id)
    return sorted(ids)


async def test_invoke() -> None:
    doc1 = Document(
        id="doc1",
        page_content="lorem ipsum and whatnot",
    )
    doc2 = Document(
        id="doc2",
        page_content="lorem ipsum and some more",
        metadata={"mentions": ["doc1"]},
    )
    store = InMemoryVectorStore.from_documents([doc1, doc2], FakeEmbeddings(size=8))

    retriever = GraphRetriever(store=store, edges=[("mentions", Id())])

    assert sorted_ids(retriever.invoke("lorem")) == ["doc1", "doc2"]
    assert sorted_ids(await retriever.ainvoke("lorem")) == [
        "doc1",
        "doc2",
    ]



"""Defines the base class for vector store adapters."""

import abc
from collections.abc import Sequence
from typing import Any, Generic, TypeVar

from graph_retriever import Content
from graph_retriever.adapters import Adapter
from graph_retriever.edges import Edge, MetadataEdge
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.runnables import run_in_executor
from langchain_core.vectorstores.base import VectorStore
from typing_extensions import (
    override,
)

from langchain_graph_retriever._conversion import doc_to_content
from langchain_graph_retriever.transformers import ShreddingTransformer

StoreT = TypeVar("StoreT", bound=VectorStore)


class LangchainAdapter(Generic[StoreT], Adapter):
    """
    Base adapter for integrating vector stores with the graph retriever system.

    This class provides a foundation for custom adapters, enabling consistent
    interaction with various vector store implementations.

    Parameters
    ----------
    vector_store :
        The vector store instance.
    """

    def __init__(
        self,
        vector_store: StoreT,
    ):
        """Initialize the base adapter."""
        self.vector_store = vector_store

    @property
    def _safe_embedding(self) -> Embeddings:
        if not self.vector_store.embeddings:
            msg = "Missing embedding"
            raise ValueError(msg)
        return self.vector_store.embeddings

    def embed_query(self, query: str):
        """Return the embedding of the query."""
        return self._safe_embedding.embed_query(query)

    async def aembed_query(self, query: str):
        """Return the embedding of the query."""
        return await self._safe_embedding.aembed_query(query)

    def update_filter_hook(
        self, filter: dict[str, Any] | None
    ) -> dict[str, Any] | None:
        """
        Update the metadata filter before executing the query.

        Parameters
        ----------
        filter :
            Filter on the metadata to update.

        Returns
        -------
        :
            The updated filter on the metadata to apply.
        """
        return filter

    def format_documents_hook(self, docs: list[Document]) -> list[Content]:
        """
        Format the documents as content after executing the query.

        Parameters
        ----------
        docs :
            The documents returned from the vector store

        Returns
        -------
        :
            The formatted content.
        """
        return [doc_to_content(doc) for doc in docs]

    @override
    def search_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[list[float], list[Content]]:
        query_embedding = self.embed_query(query)
        docs = self.search(
            embedding=query_embedding,
            k=k,
            filter=filter,
            **kwargs,
        )
        return query_embedding, docs

    @override
    async def asearch_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[list[float], list[Content]]:
        query_embedding = await self.aembed_query(query)
        docs = await self.asearch(
            embedding=query_embedding,
            k=k,
            filter=filter,
            **kwargs,
        )
        return query_embedding, docs

    @abc.abstractmethod
    def _search(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        """
        Return docs (with embeddings) most similar to the query vector.

        Parameters
        ----------
        embedding :
            Embedding to look up documents similar to.
        k :
            Number of Documents to return.
        filter :
            Filter on the metadata to apply.
        kwargs :
            Additional keyword arguments.

        Returns
        -------
        :
            List of Documents most similar to the query vector.

            Documents should have their embedding added to the
            metadata under the `METADATA_EMBEDDING_KEY` key.
        """

    @override
    def search(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        """
        Return contents most similar to the query vector.

        Parameters
        ----------
        embedding :
            Embedding to look up documents similar to.
        k :
            Number of Documents to return.
        filter :
            Filter on the metadata to apply.
        kwargs :
            Additional keyword arguments.

        Returns
        -------
        :
            List of Contents most similar to the query vector.
        """
        if k == 0:
            return []

        docs = self._search(
            embedding=embedding,
            k=k,
            filter=self.update_filter_hook(filter),
            **kwargs,
        )
        return self.format_documents_hook(docs)

    async def _asearch(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        """
        Asynchronously return docs (with embeddings) most similar to the query vector.

        Parameters
        ----------
        embedding :
            Embedding to look up documents similar to.
        k :
            Number of Documents to return.
        filter :
            Filter on the metadata to apply.
        kwargs :
            Additional keyword arguments.

        Returns
        -------
        :
            List of Documents most similar to the query vector.

            Documents should have their embedding added to the
            metadata under the `METADATA_EMBEDDING_KEY` key.
        """
        return await run_in_executor(
            None,
            self._search,
            embedding,
            k,
            filter,
            **kwargs,
        )

    @override
    async def asearch(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        if k == 0:
            return []

        docs = await self._asearch(
            embedding=embedding,
            k=k,
            filter=self.update_filter_hook(filter),
            **kwargs,
        )
        return self.format_documents_hook(docs)

    def _remove_duplicates(self, ids: Sequence[str]) -> list[str]:
        """
        Remove duplicate ids while preserving order.

        Parameters
        ----------
        ids :
            List of IDs to get.

        Returns
        -------
        :
            List of IDs with duplicates removed
        """
        return list({k: True for k in ids}.keys())

    @override
    def get(
        self,
        ids: Sequence[str],
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        docs = self._get(self._remove_duplicates(ids), filter, **kwargs)
        return self.format_documents_hook(docs)

    @abc.abstractmethod
    def _get(
        self,
        ids: Sequence[str],
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        """
        Get documents by ID.

        Fewer documents may be returned than requested if some IDs are not found
        or if there are duplicated IDs. This method should **NOT** raise
        exceptions if no documents are found for some IDs.

        Users should not assume that the order of the returned documents matches
        the order of the input IDs. Instead, users should rely on the ID field
        of the returned documents.

        Parameters
        ----------
        ids :
            List of IDs to get.
        filter :
            Filter to apply to the recrods.
        kwargs :
            Additional keyword arguments. These are up to the implementation.

        Returns
        -------
        :
            List of documents that were found.
        """

    @override
    async def aget(
        self,
        ids: Sequence[str],
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        docs = await self._aget(self._remove_duplicates(ids), filter, **kwargs)
        return self.format_documents_hook(docs)

    async def _aget(
        self,
        ids: Sequence[str],
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        """
        Asynchronously get documents by ID.

        Fewer documents may be returned than requested if some IDs are not found
        or if there are duplicated IDs. This method should **NOT** raise
        exceptions if no documents are found for some IDs.

        Users should not assume that the order of the returned documents matches
        the order of the input IDs. Instead, users should rely on the ID field
        of the returned documents.

        Parameters
        ----------
        ids :
            List of IDs to get.
        filter :
            Filter to apply to the documents.
        kwargs :
            Additional keyword arguments. These are up to the implementation.

        Returns
        -------
        :
            List of documents that were found.
        """
        return await run_in_executor(
            None,
            self._get,
            ids,
            filter,
            **kwargs,
        )

    def _metadata_filter(
        self,
        base_filter: dict[str, Any] | None = None,
        edge: Edge | None = None,
    ) -> dict[str, Any]:
        """
        Return a filter for the `base_filter` and incoming edges from `edge`.

        Parameters
        ----------
        base_filter :
            Any base metadata filter that should be used for search.
            Generally corresponds to the user specified filters for the entire
            traversal. Should be combined with the filters necessary to support
            nodes with an *incoming* edge matching `edge`.
        edge :
            An optional edge which should be added to the filter.

        Returns
        -------
        :
            The metadata dictionary to use for the given filter.
        """
        metadata_filter = {**(base_filter or {})}
        assert isinstance(edge, MetadataEdge)
        if edge is None:
            metadata_filter
        else:
            metadata_filter[edge.incoming_field] = edge.value
        return metadata_filter


class ShreddedLangchainAdapter(LangchainAdapter[StoreT]):
    """
    Base adapter for integrating vector stores with the graph retriever system.

    This class provides a foundation for custom adapters, enabling consistent
    interaction with various vector store implementations that do not support
    searching on list-based metadata values.

    Parameters
    ----------
    vector_store :
        The vector store instance.
    shredder: ShreddingTransformer, optional
        An instance of the ShreddingTransformer used for doc insertion.
        If not passed then a default instance of ShreddingTransformer is used.
    nested_metadata_fields: set[str]
        The set of metadata fields that contain nested values.
    """

    def __init__(
        self,
        vector_store: StoreT,
        shredder: ShreddingTransformer | None = None,
        nested_metadata_fields: set[str] = set(),
    ):
        """Initialize the base adapter."""
        super().__init__(vector_store=vector_store)
        self.shredder = ShreddingTransformer() if shredder is None else shredder
        self.nested_metadata_fields = nested_metadata_fields

    @override
    def update_filter_hook(
        self, filter: dict[str, str] | None
    ) -> dict[str, str] | None:
        if filter is None:
            return None

        shredded_filter = {}
        for key, value in filter.items():
            if key in self.nested_metadata_fields:
                shredded_filter[self.shredder.shredded_key(key, value)] = (
                    self.shredder.shredded_value()
                )
            else:
                shredded_filter[key] = value
        return shredded_filter

    @override
    def format_documents_hook(self, docs: list[Document]) -> list[Content]:
        restored = list(self.shredder.restore_documents(documents=docs))
        return super().format_documents_hook(restored)
